from __future__ import annotations

import logging
from contextlib import suppress
from functools import reduce
from queue import LifoQueue
from typing import TYPE_CHECKING, Iterable

from semantic_release.commit_parser import ParsedCommit
from semantic_release.commit_parser.token import ParseError
from semantic_release.const import DEFAULT_VERSION
from semantic_release.enums import LevelBump, SemanticReleaseLogLevels
from semantic_release.errors import InternalError, InvalidVersion
from semantic_release.globals import logger
from semantic_release.helpers import validate_types_in_sequence

if TYPE_CHECKING:  # pragma: no cover
    from typing import Sequence

    from git.objects.commit import Commit
    from git.refs.tag import Tag
    from git.repo.base import Repo

    from semantic_release.commit_parser import (
        CommitParser,
        ParseResult,
        ParserOptions,
    )
    from semantic_release.version.translator import VersionTranslator
    from semantic_release.version.version import Version


def tags_and_versions(
    tags: Iterable[Tag], translator: VersionTranslator
) -> list[tuple[Tag, Version]]:
    """
    Return a list of 2-tuples, where each element is a tuple (tag, version)
    from the tags in the Git repo and their corresponding `Version` according
    to `Version.from_tag`. The returned list is sorted according to semver
    ordering rules.

    Tags which are not matched by `translator` are ignored.
    """
    ts_and_vs: list[tuple[Tag, Version]] = []
    for tag in tags:
        try:
            version = translator.from_tag(tag.name)
        except (NotImplementedError, InvalidVersion) as e:
            logger.warning(
                "Couldn't parse tag %s as as Version: %s",
                tag.name,
                str(e),
                exc_info=logger.isEnabledFor(logging.DEBUG),
            )
            continue

        if version:
            ts_and_vs.append((tag, version))

    logger.info("found %s previous tags", len(ts_and_vs))
    return sorted(ts_and_vs, reverse=True, key=lambda v: v[1])


def _traverse_graph_for_commits(
    head_commit: Commit,
    latest_release_tag_str: str = "",
) -> Sequence[Commit]:
    # Depth-first search
    def dfs(start_commit: Commit, stop_nodes: set[Commit]) -> Sequence[Commit]:
        # Create a stack for DFS
        stack: LifoQueue[Commit] = LifoQueue()

        # Create a set to store visited graph nodes (commit objects in this case)
        visited: set[Commit] = set()

        # Initialize the result
        commits: list[Commit] = []

        # Add the source node in the queue to start the search
        stack.put(start_commit)

        # Traverse the git history capturing each commit found before it reaches a stop node
        while not stack.empty():
            if (node := stack.get()) in visited or node in stop_nodes:
                continue

            visited.add(node)
            commits.append(node)

            # Add all parent commits to the stack from left to right so that the rightmost is popped first
            # as the left side is generally the merged into branch
            for parent in node.parents:
                stack.put(parent)

        return commits

    # Run a Depth First Search to find all the commits since the last release
    return dfs(
        start_commit=head_commit,
        stop_nodes=set(
            head_commit.repo.iter_commits(latest_release_tag_str)
            if latest_release_tag_str
            else []
        ),
    )


def _increment_version(
    latest_version: Version,
    latest_full_version: Version,
    level_bump: LevelBump,
    prerelease: bool,
    prerelease_token: str,
    major_on_zero: bool,
    allow_zero_version: bool,
) -> Version:
    """
    Using the given versions, along with a given `level_bump`, increment to
    the next version according to whether or not this is a prerelease.

    `latest_version` is the most recent version released from this branch's history.
    `latest_full_version`, the most recent full release (i.e. not a prerelease)
    in this branch's history.

    `latest_version` and `latest_full_version` can be the same, but aren't necessarily.
    """
    local_vars = list(locals().items())
    logger.log(
        SemanticReleaseLogLevels.SILLY,
        "_increment_version: %s",
        str.join(", ", [f"{k} = {v}" for k, v in local_vars]),
    )

    # Handle variations where the latest version is 0.x.x
    if latest_version.major == 0:
        if not allow_zero_version:
            # Set up default version to be 1.0.0 if currently 0.x.x which means a commented
            # breaking change is not required to bump to 1.0.0
            logger.debug(
                "Bumping major version as 0.x.x versions are disabled because of allow_zero_version=False"
            )
            level_bump = LevelBump.MAJOR

        elif not major_on_zero:
            # if we are a 0.x.y release and have set `major_on_zero`,
            # breaking changes should increment the minor digit
            # Correspondingly, we reduce the level that we increment the
            # version by.
            logger.debug(
                "reducing version increment due to 0. version and major_on_zero=False"
            )

            level_bump = min(level_bump, LevelBump.MINOR)

    logger.debug(
        "prerelease=%s and the latest version %s %s prerelease",
        prerelease,
        latest_version,
        "is a" if latest_version.is_prerelease else "is not a",
    )

    if level_bump == LevelBump.NO_RELEASE:
        raise ValueError("level_bump must be at least PRERELEASE_REVISION")

    if level_bump == LevelBump.PRERELEASE_REVISION and not latest_version.is_prerelease:
        raise ValueError(
            "Cannot increment a non-prerelease version with a prerelease level bump"
        )

    # assume we always want to increment the version that is the latest in the branch's history
    base_version = latest_version

    # if the current version is a prerelease & we want a new prerelease, then
    # figure out if we need to bump the prerelease revision or start a new prerelease
    if latest_version.is_prerelease:
        # find the change since the last full release because if the current version is a prerelease
        # then we need to predict properly the next full version
        diff_with_last_released_version = latest_version - latest_full_version
        logger.debug(
            "the diff b/w the latest version '%s' and the latest full release version '%s' is: %s",
            latest_version,
            latest_full_version,
            diff_with_last_released_version,
        )

        # Since the difference is less than or equal to the level bump and we want a new prerelease,
        # we can abort early and just increment the revision
        if level_bump <= diff_with_last_released_version:
            # 6a ii) if level_bump <= the level bump introduced by the previous tag (latest_version)
            if prerelease:
                logger.debug(
                    "there has already been at least a %s release since the last full release %s",
                    level_bump,
                    latest_full_version,
                )
                logger.debug("Incrementing the prerelease revision...")
                new_revision = base_version.to_prerelease(
                    token=prerelease_token,
                    revision=(
                        1
                        if latest_version.prerelease_token != prerelease_token
                        else (latest_version.prerelease_revision or 0) + 1
                    ),
                )
                logger.debug("Incremented %s to %s", base_version, new_revision)
                return new_revision

            # When we don't want a prerelease, but the previous version is a prerelease that
            # had a greater bump than we currently are applying, choose the larger bump instead
            # as it consumes this bump
            logger.debug("Finalizing the prerelease version...")
            return base_version.finalize_version()

        # Fallthrough to handle all larger level bumps
        logger.debug(
            "this release has a greater bump than any change since the last full release, %s",
            latest_full_version,
        )

        # Fallthrough, if we don't want a prerelease, or if we do but the level bump is greater
        #
        # because the current version is a prerelease, we must start from the last full version
        # Case 1: we identified that the level bump is greater than the change since
        #         the last full release, this will also reset the prerelease revision
        # Case 2: we don't want a prerelease, so consider only the last full version in history
        base_version = latest_full_version

    # From the base version, we can now increment the version according to the level bump
    # regardless of the prerelease status as bump() handles the reset and pass through
    logger.debug("Bumping %s with a %s bump", base_version, level_bump)
    target_next_version = base_version.bump(level_bump)

    # Converting to/from a prerelease if necessary
    target_next_version = (
        target_next_version.to_prerelease(token=prerelease_token)
        if prerelease
        else target_next_version.finalize_version()
    )

    logger.debug("Incremented %s to %s", base_version, target_next_version)
    return target_next_version


def next_version(
    repo: Repo,
    translator: VersionTranslator,
    commit_parser: CommitParser[ParseResult, ParserOptions],
    allow_zero_version: bool,
    major_on_zero: bool,
    prerelease: bool = False,
) -> Version:
    """
    Evaluate the history within `repo`, and based on the tags and commits in the repo
    history, identify the next semantic version that should be applied to a release
    """
    # Default initial version
    # Since the translator is configured by the user, we can't guarantee that it will
    # be able to parse the default version. So we first cast it to a tag using the default
    # value and the users configured tag format, then parse it back to a version object
    default_initial_version = translator.from_tag(
        translator.str_to_tag(DEFAULT_VERSION)
    )
    if default_initial_version is None:
        # This should never happen, but if it does, it's a bug
        raise InternalError(
            "Translator was unable to parse the embedded default version"
        )

    # Step 1. All tags, sorted descending by semver ordering rules
    all_git_tags_as_versions = tags_and_versions(repo.tags, translator)

    # Retrieve all commit hashes (regardless of merges) in the current branch's history from repo origin
    commit_hash_set = {
        commit.hexsha
        for commit in _traverse_graph_for_commits(head_commit=repo.active_branch.commit)
    }

    # Filter all releases that are not found in the current branch's history
    historic_versions: list[Version] = []
    for tag, version in all_git_tags_as_versions:
        # TODO: move this to tags_and_versions() function?
        # Ignore the error that is raised when tag points to a Blob or Tree object rather
        # than a commit object (tags that point to tags that then point to commits are resolved automatically)
        with suppress(ValueError):
            if tag.commit.hexsha in commit_hash_set:
                historic_versions.append(version)

    # Step 2. Get the latest final release version in the history of the current branch
    #  or fallback to the default 0.0.0 starting version value if none are found
    latest_full_release_version = next(
        filter(
            lambda version: not version.is_prerelease,
            historic_versions,
        ),
        default_initial_version,
    )

    logger.info(
        f"The last full version in this branch's history was {latest_full_release_version}"
        if latest_full_release_version != default_initial_version
        else "No full releases found in this branch's history"
    )

    # Step 3. Determine the latest release version in the history of the current branch
    # If we the desired result is a prerelease, we must determine if there was any previous
    # prerelease in the history of the current branch beyond the latest_full_release_version.
    # Important to note that, we only consider prereleases that are of the same prerelease token
    # as the basis of incrementing the prerelease revision.
    # If we are not looking for a prerelease, this is the same as the last full release.
    latest_version = (
        latest_full_release_version
        if not prerelease
        else next(
            filter(
                lambda version: all(
                    [
                        version.is_prerelease,
                        version.prerelease_token == translator.prerelease_token,
                        version >= latest_full_release_version,
                    ]
                ),
                historic_versions,
            ),
            latest_full_release_version,  # default
        )
    )

    logger.info("The latest release in this branch's history was %s", latest_version)

    # Step 4. Walk the git tree to find all commits that have been made since the last release
    commits_since_last_release = _traverse_graph_for_commits(
        head_commit=repo.active_branch.commit,
        latest_release_tag_str=(
            # NOTE: the default_initial_version should not actually exist on the repository (ie v0.0.0)
            # so we provide an empty tag string when there are no tags on the repository yet
            latest_version.as_tag() if latest_version != default_initial_version else ""
        ),
    )

    logger.info(
        f"Found {len(commits_since_last_release)} commits since the last release!"
        if len(commits_since_last_release) > 0
        else "No commits found since the last release!"
    )

    # Step 5. apply the parser to each commit in the history (could return multiple results per commit)
    parsed_results = list(map(commit_parser.parse, commits_since_last_release))

    # Step 5A. Accumulate all parsed results into a single list accounting for possible multiple results per commit
    consolidated_results: list[ParseResult] = reduce(
        lambda accumulated_results, p_results: [
            *accumulated_results,
            *(
                # Cast to list if not already a list
                p_results
                if isinstance(p_results, list) or type(p_results) == tuple
                else [p_results]
            ),
        ],
        parsed_results,
        [],
    )

    # Step 5B. Validation type check for the parser results (important because of possible custom parsers)
    if not validate_types_in_sequence(consolidated_results, (ParseError, ParsedCommit)):
        raise TypeError("Unexpected type returned from commit_parser.parse")

    # Step 5C. Parse the commits to determine the bump level that should be applied
    parsed_levels: set[LevelBump] = {
        parsed_result.bump  # type: ignore[union-attr] # too complex for type checkers
        for parsed_result in filter(
            # Filter out any non-ParsedCommit results (i.e. ParseErrors)
            lambda parsed_result: isinstance(parsed_result, ParsedCommit),
            consolidated_results,
        )
    }

    logger.debug(
        "parsed the following distinct levels from the commits since the last release: %s",
        parsed_levels,
    )

    level_bump = max(parsed_levels, default=LevelBump.NO_RELEASE)
    logger.info("The type of the next release release is: %s", level_bump)

    if all(
        [
            level_bump is LevelBump.NO_RELEASE,
            latest_version.major != 0 or allow_zero_version,
        ]
    ):
        logger.info("No release will be made")
        return latest_version

    return _increment_version(
        latest_version=latest_version,
        latest_full_version=latest_full_release_version,
        level_bump=level_bump,
        prerelease=prerelease,
        prerelease_token=translator.prerelease_token,
        major_on_zero=major_on_zero,
        allow_zero_version=allow_zero_version,
    )



from __future__ import annotations

from contextlib import suppress
from copy import deepcopy
from functools import reduce
from re import MULTILINE, compile as regexp
from typing import TYPE_CHECKING

# TODO: remove in v11
from semantic_release.helpers import (
    sort_numerically,  # noqa: F401 # TODO: maintained for compatibility
)

if TYPE_CHECKING:  # pragma: no cover
    from re import Pattern
    from typing import Any, TypedDict

    from git import Commit

    class RegexReplaceDef(TypedDict):
        pattern: Pattern
        repl: str


breaking_re = regexp(r"BREAKING[ -]CHANGE:\s?(.*)")

un_word_wrap: RegexReplaceDef = {
    # Match a line ending where the next line is not indented, or a bullet
    "pattern": regexp(r"((?<!-)\n(?![\s*-]))"),
    "repl": r" ",  # Replace with a space
}

un_word_wrap_hyphen: RegexReplaceDef = {
    "pattern": regexp(r"((?<=\w)-\n(?=\w))"),
    "repl": r"-",  # Replace with single hyphen
}

trim_line_endings: RegexReplaceDef = {
    # Match line endings with optional whitespace
    "pattern": regexp(r"[\r\t\f\v ]*\r?\n"),
    "repl": "\n",  # remove the optional whitespace & remove windows newlines
}

spread_out_git_footers: RegexReplaceDef = {
    # Match a git footer line, and add an extra newline after it
    # only be flexible enough for a double space indent (otherwise its probably on purpose)
    #   - found collision with dependabot's yaml in a commit message with a git footer (its unusal but possible)
    "pattern": regexp(r"^ {0,2}([\w-]*: .+)$\n?((?!\n) *[^:\n]+:)", MULTILINE),
    "repl": r"\1\n\n\2",
}


def parse_paragraphs(text: str) -> list[str]:
    r"""
    This will take a text block and return a list containing each
    paragraph with single line breaks collapsed into spaces.

    To handle Windows line endings, carriage returns '\r' are removed before
    separating into paragraphs.

    It will attempt to detect Git footers and they will not be condensed.

    :param text: The text string to be divided.
    :return: A list of condensed paragraphs, as strings.
    """
    adjusted_text = reduce(
        lambda txt, adj: adj["pattern"].sub(adj["repl"], txt),
        [trim_line_endings, un_word_wrap_hyphen],
        text,
    )

    # Repeat replacements until no more changes are made
    prev_iteration = ""
    while prev_iteration != adjusted_text:
        prev_iteration = adjusted_text
        adjusted_text = spread_out_git_footers["pattern"].sub(
            spread_out_git_footers["repl"], adjusted_text
        )

    return list(
        filter(
            None,
            [
                un_word_wrap["pattern"].sub(un_word_wrap["repl"], paragraph).strip()
                for paragraph in adjusted_text.strip().split("\n\n")
            ],
        )
    )


def force_str(msg: str | bytes | bytearray | memoryview) -> str:
    # This shouldn't be a thing but typing is being weird around what
    # git.commit.message returns and the memoryview type won't go away
    message = msg.tobytes() if isinstance(msg, memoryview) else msg
    return (
        message.decode("utf-8")
        if isinstance(message, (bytes, bytearray))
        else str(message)
    )


def deep_copy_commit(commit: Commit) -> dict[str, Any]:
    keys = [
        "repo",
        "binsha",
        "author",
        "authored_date",
        "committer",
        "committed_date",
        "message",
        "tree",
        "parents",
        "encoding",
        "gpgsig",
        "author_tz_offset",
        "committer_tz_offset",
    ]
    kwargs = {}
    for key in keys:
        with suppress(ValueError):
            if hasattr(commit, key) and (value := getattr(commit, key)) is not None:
                if key in ["parents", "repo", "tree"]:
                    # These tend to have circular references so don't deepcopy them
                    kwargs[key] = value
                    continue

                kwargs[key] = deepcopy(value)

    return kwargs



from __future__ import annotations

from itertools import chain, zip_longest
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from semantic_release.commit_parser.scipy import ScipyCommitParser
from semantic_release.enums import LevelBump

if TYPE_CHECKING:
    from typing import Protocol

    from semantic_release.commit_parser.scipy import ScipyParserOptions

    class FormatScipyCommitFn(Protocol):
        def __call__(
            self, scipy_tag: str, subject: str, body_parts: list[str]
        ) -> str: ...


@pytest.fixture(scope="session")
def format_scipy_commit():
    def _format_scipy_commit(
        scipy_tag: str, subject: str, body_parts: list[str]
    ) -> str:
        body = str.join("\n\n", body_parts)
        return f"{scipy_tag}: {subject}\n\n{body}"

    return _format_scipy_commit


@pytest.fixture(scope="session")
def default_scipy_parser() -> ScipyCommitParser:
    return ScipyCommitParser()


@pytest.fixture(scope="session")
def default_scipy_parser_options(
    default_scipy_parser: ScipyCommitParser,
) -> ScipyParserOptions:
    return default_scipy_parser.get_default_options()


@pytest.fixture(scope="session")
def scipy_chore_commit_types(
    default_scipy_parser_options: ScipyParserOptions,
) -> list[str]:
    return [
        k
        for k, v in default_scipy_parser_options.tag_to_level.items()
        if v < LevelBump.PATCH
    ]


@pytest.fixture(scope="session")
def scipy_patch_commit_types(
    default_scipy_parser_options: ScipyParserOptions,
) -> list[str]:
    return [
        k
        for k, v in default_scipy_parser_options.tag_to_level.items()
        if v is LevelBump.PATCH
    ]


@pytest.fixture(scope="session")
def scipy_minor_commit_types(
    default_scipy_parser_options: ScipyParserOptions,
) -> list[str]:
    return [
        k
        for k, v in default_scipy_parser_options.tag_to_level.items()
        if v is LevelBump.MINOR
    ]


@pytest.fixture(scope="session")
def scipy_major_commit_types(
    default_scipy_parser_options: ScipyParserOptions,
) -> list[str]:
    return [
        k
        for k, v in default_scipy_parser_options.tag_to_level.items()
        if v is LevelBump.MAJOR
    ]


@pytest.fixture(scope="session")
def scipy_nonparseable_commits() -> list[str]:
    return [
        "Initial Commit",
        "Merge pull request #14447 from AnirudhDagar/rename_ndimage_modules",
    ]


@pytest.fixture(scope="session")
def scipy_chore_subjects(scipy_chore_commit_types: list[str]) -> list[str]:
    subjects = {
        "BENCH": "disable very slow benchmark in optimize_milp.py",
        "DEV": "add unicode check to pre-commit-hook",
        "DOC": "change approx_fprime doctest (#20568)",
        "STY": "fixed ruff & mypy issues",
        "TST": "Skip Cython tests for editable installs",
        "REL": "set version to 1.0.0",
        "TEST": "Add Cython tests for editable installs",
    }
    # Test fixture modification failure prevention
    assert len(subjects.keys()) == len(scipy_chore_commit_types)
    return [subjects[k] for k in scipy_chore_commit_types]


@pytest.fixture(scope="session")
def scipy_patch_subjects(scipy_patch_commit_types: list[str]) -> list[str]:
    subjects = {
        "BLD": "move the optimize build steps earlier into the build sequence",
        "BUG": "Fix invalid default bracket selection in _bracket_minimum (#20563)",
        "MAINT": "optimize.linprog: fix bug when integrality is a list of all zeros (#20586)",
    }
    # Test fixture modification failure prevention
    assert len(subjects.keys()) == len(scipy_patch_commit_types)
    return [subjects[k] for k in scipy_patch_commit_types]


@pytest.fixture(scope="session")
def scipy_minor_subjects(scipy_minor_commit_types: list[str]) -> list[str]:
    subjects = {
        "ENH": "stats.ttest_1samp: add array-API support (#20545)",
        # "REV": "reverted a previous commit",
        "FEAT": "added a new feature",
    }
    # Test fixture modification failure prevention
    assert len(subjects.keys()) == len(scipy_minor_commit_types)
    return [subjects[k] for k in scipy_minor_commit_types]


@pytest.fixture(scope="session")
def scipy_major_subjects(scipy_major_commit_types: list[str]) -> list[str]:
    subjects = {
        "API": "dropped support for python 3.7",
        "DEP": "stats: switch kendalltau to kwarg-only, remove initial_lexsort",
    }
    # Test fixture modification failure prevention
    assert len(subjects.keys()) == len(scipy_major_commit_types)
    return [subjects[k] for k in scipy_major_commit_types]


@pytest.fixture(scope="session")
def scipy_brk_change_commit_bodies() -> list[list[str]]:
    brk_chg_msg = dedent(
        """
        BREAKING CHANGE: a description of what is now different
        with multiple lines
        """
    ).strip()

    one_line_desc = "resolves bug related to windows incompatiblity"

    return [
        # No regular change description
        [brk_chg_msg],
        # regular change description & breaking change message
        [one_line_desc, brk_chg_msg],
        # regular change description & breaking change message with footer
        [one_line_desc, brk_chg_msg, "Resolves: #666"],
    ]


@pytest.fixture(scope="session")
def scipy_nonbrking_commit_bodies() -> list[list[str]]:
    # a GitHub squash merge that preserved PR commit messages (all chore-like)
    github_squash_merge_body = str.join(
        "\n\n",
        [
            "* DOC: import ropy.transform to test for numpy error",
            "* DOC: lower numpy version",
            "* DOC: lower numpy version further",
            "* STY: resolve linting issues",
        ],
    )

    one_block_desc = dedent(
        """
        Bug spotted on Fedora, see https://src.fedoraproject.org/rpms/scipy/pull-request/22
        with an additional multiline description
        """
    ).strip()

    return [
        github_squash_merge_body.split("\n\n"),  # split into blocks
        # empty body
        [],
        [""],
        # formatted body (ie dependabot)
        dedent(
            """
            Bumps [package](https://github.com/namespace/project) from 3.5.3 to 4.1.1.
                - [Release notes](https://github.com/namespace/project/releases)
                - [Changelog](https://github.com/namespace/project/blob/4.x/CHANGES)
                - [Commits](https://github.com/namespace/project/commits/v4.1.1)

            ---
            updated-dependencies:
              - dependency-name: package
                dependency-type: direct:development
                update-type: version-update:semver-major
            """
        )
        .lstrip()
        .split("\n\n"),
        # 1 block description
        one_block_desc.split("\n\n"),
        # keywords
        ["[skip azp] [skip actions]"],
        # Resolving an issue on GitHub
        ["Resolves: #127"],
        [one_block_desc, "Closes: #1024"],
    ]


@pytest.fixture(scope="session")
def scipy_chore_commit_parts(
    scipy_chore_commit_types: list[str],
    scipy_chore_subjects: list[str],
    scipy_nonbrking_commit_bodies: list[list[str]],
) -> list[tuple[str, str, list[str]]]:
    # Test fixture modification failure prevention
    assert len(scipy_chore_commit_types) == len(scipy_chore_subjects)

    # build full commit messages with commit type prefix, subject, and body variant
    # for all body variants
    return [
        (commit_type, subject, commit_body_blocks)
        for commit_type, subject in zip(scipy_chore_commit_types, scipy_chore_subjects)
        for commit_body_blocks in scipy_nonbrking_commit_bodies
    ]


@pytest.fixture(scope="session")
def scipy_chore_commits(
    scipy_chore_commit_parts: list[tuple[str, str, list[str]]],
    format_scipy_commit: FormatScipyCommitFn,
) -> list[str]:
    # build full commit messages with commit type prefix, subject, and body variant
    # for all body variants
    return [
        format_scipy_commit(commit_type, subject, commit_body)
        for commit_type, subject, commit_body in scipy_chore_commit_parts
    ]


@pytest.fixture(scope="session")
def scipy_patch_commit_parts(
    scipy_patch_commit_types: list[str],
    scipy_patch_subjects: list[str],
    scipy_nonbrking_commit_bodies: list[list[str]],
) -> list[tuple[str, str, list[str]]]:
    # Test fixture modification failure prevention
    assert len(scipy_patch_commit_types) == len(scipy_patch_subjects)

    # build full commit messages with commit type prefix, subject, and body variant
    # for all body variants
    return [
        (commit_type, subject, commit_body_blocks)
        for commit_type, subject in zip(scipy_patch_commit_types, scipy_patch_subjects)
        for commit_body_blocks in scipy_nonbrking_commit_bodies
    ]


@pytest.fixture(scope="session")
def scipy_patch_commits(
    scipy_patch_commit_parts: list[tuple[str, str, list[str]]],
    format_scipy_commit: FormatScipyCommitFn,
) -> list[str]:
    # build full commit messages with commit type prefix, subject, and body variant
    # for all body variants
    return [
        format_scipy_commit(commit_type, subject, commit_body)
        for commit_type, subject, commit_body in scipy_patch_commit_parts
    ]


@pytest.fixture(scope="session")
def scipy_minor_commit_parts(
    scipy_minor_commit_types: list[str],
    scipy_minor_subjects: list[str],
    scipy_nonbrking_commit_bodies: list[list[str]],
) -> list[tuple[str, str, list[str]]]:
    # Test fixture modification failure prevention
    assert len(scipy_minor_commit_types) == len(scipy_minor_subjects)

    # build full commit messages with commit type prefix, subject, and body variant
    # for all body variants
    return [
        (commit_type, subject, commit_body_blocks)
        for commit_type, subject in zip(scipy_minor_commit_types, scipy_minor_subjects)
        for commit_body_blocks in scipy_nonbrking_commit_bodies
    ]


@pytest.fixture(scope="session")
def scipy_minor_commits(
    scipy_minor_commit_parts: list[tuple[str, str, list[str]]],
    format_scipy_commit: FormatScipyCommitFn,
) -> list[str]:
    # build full commit messages with commit type prefix, subject, and body variant
    # for all body variants
    return [
        format_scipy_commit(commit_type, subject, commit_body)
        for commit_type, subject, commit_body in scipy_minor_commit_parts
    ]


@pytest.fixture(scope="session")
def scipy_major_commit_parts(
    scipy_major_commit_types: list[str],
    scipy_major_subjects: list[str],
    scipy_brk_change_commit_bodies: list[list[str]],
) -> list[tuple[str, str, list[str]]]:
    # Test fixture modification failure prevention
    assert len(scipy_major_commit_types) == len(scipy_major_subjects)

    # build full commit messages with commit type prefix, subject, and body variant
    # for all body variants
    return [
        (commit_type, subject, commit_body_blocks)
        for commit_type, subject in zip(scipy_major_commit_types, scipy_major_subjects)
        for commit_body_blocks in scipy_brk_change_commit_bodies
    ]


@pytest.fixture(scope="session")
def scipy_major_commits(
    scipy_major_commit_parts: list[tuple[str, str, list[str]]],
    format_scipy_commit: FormatScipyCommitFn,
) -> list[str]:
    # build full commit messages with commit type prefix, subject, and body variant
    # for all body variants
    return [
        format_scipy_commit(commit_type, subject, commit_body)
        for commit_type, subject, commit_body in scipy_major_commit_parts
    ]


@pytest.fixture(scope="session")
def scipy_patch_mixed_commits(
    scipy_patch_commits: list[str],
    scipy_chore_commits: list[str],
) -> list[str]:
    return list(
        filter(
            None,
            chain.from_iterable(zip_longest(scipy_patch_commits, scipy_chore_commits)),
        )
    )


@pytest.fixture(scope="session")
def scipy_minor_mixed_commits(
    scipy_minor_commits: list[str],
    scipy_patch_commits: list[str],
    scipy_chore_commits: list[str],
) -> list[str]:
    return list(
        chain.from_iterable(
            zip_longest(
                scipy_minor_commits,
                scipy_patch_commits,
                scipy_chore_commits,
                fillvalue="uninteresting",
            )
        )
    )


@pytest.fixture(scope="session")
def scipy_major_mixed_commits(
    scipy_major_commits: list[str],
    scipy_minor_commits: list[str],
    scipy_patch_commits: list[str],
    scipy_chore_commits: list[str],
) -> list[str]:
    return list(
        filter(
            None,
            chain.from_iterable(
                zip_longest(
                    scipy_major_commits,
                    scipy_minor_commits,
                    scipy_patch_commits,
                    scipy_chore_commits,
                )
            ),
        )
    )



"""Python Semantic Release"""

from __future__ import annotations

import importlib.metadata

from semantic_release.commit_parser import (
    CommitParser,
    ParsedCommit,
    ParseError,
    ParseResult,
    ParseResultType,
    ParserOptions,
)
from semantic_release.enums import LevelBump
from semantic_release.errors import (
    CommitParseError,
    InvalidConfiguration,
    InvalidVersion,
    SemanticReleaseBaseError,
)
from semantic_release.version import (
    Version,
    VersionTranslator,
    next_version,
    tags_and_versions,
)

__version__ = importlib.metadata.version(f"python_{__package__}".replace("_", "-"))

__all__ = [
    "CommitParser",
    "ParsedCommit",
    "ParseError",
    "ParseResult",
    "ParseResultType",
    "ParserOptions",
    "LevelBump",
    "SemanticReleaseBaseError",
    "CommitParseError",
    "InvalidConfiguration",
    "InvalidVersion",
    "Version",
    "VersionTranslator",
    "next_version",
    "tags_and_versions",
]


def setup_hook(argv: list[str]) -> None:
    """
    A hook to be used in setup.py to enable `python setup.py publish`.

    :param argv: sys.argv
    """
    if len(argv) > 1 and any(
        cmd in argv for cmd in ["version", "publish", "changelog"]
    ):
        from semantic_release.cli.commands.main import main

        main()



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pytest
from freezegun import freeze_time

from tests.fixtures.repos.git_flow import (
    repo_w_git_flow_conventional_commits,
    repo_w_git_flow_emoji_commits,
    repo_w_git_flow_scipy_commits,
)
from tests.util import temporary_working_directory

if TYPE_CHECKING:
    from pathlib import Path
    from unittest.mock import MagicMock

    from requests_mock import Mocker

    from semantic_release.version.version import Version

    from tests.e2e.cmd_version.bump_version.conftest import (
        InitMirrorRepo4RebuildFn,
        RunPSReleaseFn,
    )
    from tests.e2e.conftest import GetSanitizedChangelogContentFn
    from tests.fixtures.example_project import ExProjectDir
    from tests.fixtures.git_repo import (
        BuildRepoFromDefinitionFn,
        BuildSpecificRepoFn,
        CommitConvention,
        GetGitRepo4DirFn,
        RepoActionRelease,
        RepoActions,
        SplitRepoActionsByReleaseTagsFn,
    )


@pytest.mark.parametrize(
    "repo_fixture_name",
    [
        pytest.param(repo_fixture_name, marks=pytest.mark.comprehensive)
        for repo_fixture_name in [
            repo_w_git_flow_conventional_commits.__name__,
            repo_w_git_flow_emoji_commits.__name__,
            repo_w_git_flow_scipy_commits.__name__,
        ]
    ],
)
def test_gitflow_repo_rebuild_1_channel(
    repo_fixture_name: str,
    run_psr_release: RunPSReleaseFn,
    build_git_flow_repo_w_1_release_channels: BuildSpecificRepoFn,
    split_repo_actions_by_release_tags: SplitRepoActionsByReleaseTagsFn,
    init_mirror_repo_for_rebuild: InitMirrorRepo4RebuildFn,
    example_project_dir: ExProjectDir,
    git_repo_for_directory: GetGitRepo4DirFn,
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    mocked_git_fetch: MagicMock,
    mocked_git_push: MagicMock,
    post_mocker: Mocker,
    version_py_file: Path,
    get_sanitized_md_changelog_content: GetSanitizedChangelogContentFn,
    get_sanitized_rst_changelog_content: GetSanitizedChangelogContentFn,
    pyproject_toml_file: Path,
    changelog_md_file: Path,
    changelog_rst_file: Path,
):
    # build target repo into a temporary directory
    target_repo_dir = example_project_dir / repo_fixture_name
    commit_type: CommitConvention = (
        repo_fixture_name.split("commits", 1)[0].split("_")[-2]  # type: ignore[assignment]
    )
    target_repo_definition = build_git_flow_repo_w_1_release_channels(
        repo_name=repo_fixture_name,
        commit_type=commit_type,
        dest_dir=target_repo_dir,
    )
    target_git_repo = git_repo_for_directory(target_repo_dir)

    # split repo actions by release actions
    organized_steps = split_repo_actions_by_release_tags(target_repo_definition)
    configuration_steps = organized_steps.pop(None)
    unreleased_steps = organized_steps.pop("Unreleased", None)
    if unreleased_steps:
        raise ValueError("Unreleased steps found. Not Supported yet!")

    release_tags_2_steps = cast("dict[Version, list[RepoActions]]", organized_steps)

    # Create the mirror repo directory
    mirror_repo_dir = init_mirror_repo_for_rebuild(
        mirror_repo_dir=(example_project_dir / "mirror"),
        configuration_steps=configuration_steps,  # type: ignore[arg-type]
        files_to_remove=[
            changelog_md_file,
            changelog_rst_file,
        ],
    )
    mirror_git_repo = git_repo_for_directory(mirror_repo_dir)

    # rebuild repo from scratch stopping before each release tag
    for curr_version, steps in release_tags_2_steps.items():
        curr_release_tag = curr_version.as_tag()

        # make sure mocks are clear
        mocked_git_fetch.reset_mock()
        mocked_git_push.reset_mock()
        post_mocker.reset_mock()

        # Extract expected result from target repo
        target_git_repo.git.checkout(curr_release_tag, detach=True)
        expected_md_changelog_content = get_sanitized_md_changelog_content(
            repo_dir=target_repo_dir
        )
        expected_rst_changelog_content = get_sanitized_rst_changelog_content(
            repo_dir=target_repo_dir
        )
        expected_pyproject_toml_content = (
            target_repo_dir / pyproject_toml_file
        ).read_text()
        expected_version_file_content = (target_repo_dir / version_py_file).read_text()
        expected_release_commit_text = target_git_repo.head.commit.message

        # In our repo env, start building the repo from the definition
        build_repo_from_definition(
            dest_dir=mirror_repo_dir,
            repo_construction_steps=steps[:-1],  # stop before the release step
        )
        release_action_step: RepoActionRelease = steps[-1]  # type: ignore[assignment]

        # Act: run PSR on the repo instead of the RELEASE step
        with freeze_time(
            release_action_step["details"]["datetime"]
        ), temporary_working_directory(mirror_repo_dir):
            run_psr_release(
                next_version_str=release_action_step["details"]["version"],
                git_repo=mirror_git_repo,
            )

        # take measurement after running the version command
        actual_release_commit_text = mirror_git_repo.head.commit.message
        actual_pyproject_toml_content = (
            mirror_repo_dir / pyproject_toml_file
        ).read_text()
        actual_version_file_content = (mirror_repo_dir / version_py_file).read_text()
        actual_md_changelog_content = get_sanitized_md_changelog_content(
            repo_dir=mirror_repo_dir
        )
        actual_rst_changelog_content = get_sanitized_rst_changelog_content(
            repo_dir=mirror_repo_dir
        )

        # Evaluate (normal release actions should have occurred as expected)
        # ------------------------------------------------------------------
        # Make sure version file is updated
        assert expected_pyproject_toml_content == actual_pyproject_toml_content
        assert expected_version_file_content == actual_version_file_content
        # Make sure changelog is updated
        assert expected_md_changelog_content == actual_md_changelog_content
        assert expected_rst_changelog_content == actual_rst_changelog_content
        # Make sure commit is created
        assert expected_release_commit_text == actual_release_commit_text
        # Make sure tag is created
        assert curr_release_tag in [tag.name for tag in mirror_git_repo.tags]
        assert (
            mocked_git_fetch.call_count == 1
        )  # fetch called to check for remote changes
        assert mocked_git_push.call_count == 2  # 1 for commit, 1 for tag
        assert post_mocker.call_count == 1  # vcs release creation occurred
